﻿#include "common.cuh"

#include <vector>
#include <iostream>
#include <fstream>
#include <stdio.h>

#include <random>
#include <cmath>
#include <string>
#include <limits>

#include "obj_loader.cuh"
#include "svd3_cuda.h"

#include "Node.cuh"
#include "Particle.cuh"
#include "Grid.cuh"
#include "sdf.cuh"
#include "emitter.cuh"

#include <curand_kernel.h>

#include <fstream>
#include "json.hpp"


// Zalozenia:
// 1. wczytujemy dane
// 2. umieszczamy czastki na scenie uzywajac SDF i wczytanych siatek .obj
// 3. wykonujemy krok symulacji metoda MPM opisana w 
// https://disneyanimation.com/publications/a-material-point-method-for-snow-simulation/
// przy czym zrownoleglamy ja uzywajac CUDA
// 4. kolizje czastek ze scianami rozwiazujemy w oparciu o SDF 
// 5. na podstawie wynikow, co pewna liczbe krokow czasowych wykonujemy triangulację uzywajac zrownoleglonego Marching Cubes


struct MaterialProperties {
    float theta_c;
    float theta_s;
    float xi;
    float rho_0;
    float E_0;
    float nu;
    float mu_0;       // Współczynnik Lamego
    float lambda_0;   // Współczynnik Lamego

    __host__ __device__ void calculateLameCoefficients() {
        mu_0 = E_0 / (2 * (1 + nu));
        lambda_0 = E_0 * nu / ((1 - 2 * nu) * (1 + nu));

    }
};

struct TimeProperties {
    float t_0; 		// czas poczatkowy, zwykle 0
    float t; 		// czas aktualny
    float dt; 		// krok czasowy 
    float t_end;	// czas zakonczenia symulacji

// krok czasowy jest bardzo istotny

// 1. z jednej strony nalezy go dostosowac do liczby klatek, wraz z liczba krokow czasowych pomijanych przy zapisywaniu siatek wynikowych
// 2. obliczenia sa prowadzone schematem jawnym, oznacza to, że jest on bardzo wrazliwy na krok czasowy
// 3. w przypadku gdy występuje rozbieznosc wynikow do nieskonczonosci, w pierwszej kolejnosci problemu mozna szukac w za duzej wartosci kroku czasowego
// 4. w tej symulacji występuje jakaś wersja warunku CFL - nalezy pamietac, ze zwiekszenie rozdzielczosci symulacji (wiecej czastek, gestsza siatka) dla zachowania stabilnosci bedzie wymagalo zmniejszenia kroku czasowgo
};

struct SimulationProperties {
    float sticky;		// spowolnienie predkosci stycznej przy kolizji, 0.9 powinno byc ok
				// jesli czastki maja sie zatrzymac od razu po kolizji sticky powinno wynosic 0.0
				// jesli maja sie nie zatrzymowac w zwiazku z kolizja, to powinno wynosic 1.0
    float alpha; 		// wspolczynnik stosunku schamatow PIC i FLIP przy wyznaczaniu predkosci, raczej nie powinien wymagac zmian i wynosic 0.95
    float particle_diameter;	// srednica czastki, od tego zalezy ile czastek bedzie na scenie, mniejsza - wiecej powinna byc zmieniana razem z h - rozdzielczoscia siatki, 4 czastki na 1 komorke siatki wydaje sie dzialac prawidlowo
    float particle_volume;  	// Obliczany automatycznie
    
    //std::vector<Precision> snow_block_rho;    // Gęstości cząstek
};

struct GridProperties {
    float hh;				        // rozdzielczosc siatki
    float MARCHING_CUBES_EPSILON;	// tolerancja w marching cubes
    float isolevel;			        // od tego zalezy od jakiej wartości zaczna sie tworzyc trojkaty w marching cubes
                                    // wieksze - "chudszy" snieg, mniejsze - bardziej "pulchny"
                                    // jesli brakuje trojkatow, to znaczy ze nalezy ja zmniejszyc
};

struct AnimationProperties {
    int step;		// co ile krokow czasowych zapisywac wynik, nalezy dobrac tak, aby wraz z krokiem czasowym uzyskac odpowiednia liczbe klatek
};

struct PathsProperties {
    std::string base_path;
    std::string result_path;
    std::string result_name;
    //std::string path_file;
    std::string walls_path;
    std::string walls_name;
    std::vector<std::string> snow_objects_paths;
    std::vector<std::string> snow_objects_names;
    std::vector<Vector> snow_block_velocities; // Wektory prędkości postępowych
    std::vector<Vector> snow_block_rotation;  // Wektory prędkości obrotowych
    std::vector<Precision> snow_block_rho; // gestosci czastek w danym bloku

};

void LoadConfiguration(const std::string& configFilePath, MaterialProperties& materialProps, TimeProperties& timeProps, SimulationProperties& simProps, GridProperties& gridProps, AnimationProperties& animProps, PathsProperties& pathProps) {
    std::ifstream configFile(configFilePath);
    if (!configFile.is_open()) {
        std::cerr << "Nie można otworzyć pliku konfiguracyjnego: " << configFilePath << std::endl;
        return;
    }

    nlohmann::json config;
    configFile >> config;

    // Wczytaj parametry materialowe
    materialProps.theta_c = config["material"]["theta_c"];
    materialProps.theta_s = config["material"]["theta_s"];
    materialProps.xi = config["material"]["xi"];
    materialProps.rho_0 = config["material"]["rho_0"];
    materialProps.E_0 = config["material"]["E_0"];
    materialProps.nu = config["material"]["nu"];

    // Wczytaj parametry czasu
    timeProps.t_0 = config["time"]["t_0"];
    timeProps.t = config["time"]["t"];
    timeProps.dt = config["time"]["dt"];
    timeProps.t_end = config["time"]["t_end"];

    // Wczytaj parametry symulacji
    simProps.sticky = config["simulation"]["sticky"];
    simProps.alpha = config["simulation"]["alpha"];
    simProps.particle_diameter = config["simulation"]["particle_diameter"];
    simProps.particle_volume = simProps.particle_diameter * simProps.particle_diameter * simProps.particle_diameter;

    // Wczytaj parametry siatki
    gridProps.hh = config["grid"]["hh"];
    gridProps.MARCHING_CUBES_EPSILON = config["grid"]["MARCHING_CUBES_EPSILON"];
    gridProps.isolevel = config["grid"]["isolevel"];

    //Inne parametry
    //otherProps.isolevel = config["other"]["isolevel"];

    //Parametry Animacji
    animProps.step = config["animation"]["step"];

    //Ścieżki do plików
    pathProps.base_path = config["base_path"].get<std::string>();
    pathProps.walls_path = pathProps.base_path + config["paths"]["walls_path"].get<std::string>();
    pathProps.walls_name = config["paths"]["walls_name"].get<std::string>();
    pathProps.result_path = pathProps.base_path + config["paths"]["result_path"].get<std::string>();
    pathProps.result_name = config["paths"]["result_name"].get<std::string>();
    pathProps.snow_objects_paths = config["paths"]["snow_objects_paths"].get<std::vector<std::string>>();
    pathProps.snow_objects_names = config["paths"]["snow_objects_names"].get<std::vector<std::string>>();


    for (int i = 0; i < pathProps.snow_objects_paths.size(); i++) { // zeby utrzymac konwencje z pathProps.result_path
        pathProps.snow_objects_paths[i] = pathProps.base_path + pathProps.snow_objects_paths[i];
    }

    for (const auto& vec : config["paths"]["snow_block_velocities"]) {
        pathProps.snow_block_velocities.emplace_back(vec[0], vec[1], vec[2]);
    }

    for (const auto& vec : config["paths"]["snow_block_rotation"]) {

        pathProps.snow_block_rotation.emplace_back(vec[0], vec[1], vec[2]);
    }

    for (const auto& rho : config["paths"]["snow_block_rho"]) {

        pathProps.snow_block_rho.emplace_back(rho);
    }
}

__device__ float device_MCE;
__device__ float device_sticky;

//__device__ Precision h;

// do gridu rzeczy
// te ponizej sa wyznaczane po wczytaniu siatki modelu (scian) wyznaczaja bounding box
__device__ Precision x_min_grid;
__device__ Precision y_min_grid;
__device__ Precision z_min_grid;

__device__ Precision x_max_grid;
__device__ Precision y_max_grid;
__device__ Precision z_max_grid;

// funkcja interpolujaca
__device__ Precision N(Precision x) {

    x = fabs(x);
    Precision value;
    if (x < 1.0) {
        value = 0.5 * x * x * x - x * x + 2.0 / 3.0;
    }
    else if (x < 2.0) {
        value = -1.0 / 6.0 * x * x * x + x * x - 2 * x + 4.0 / 3.0;
    }
    else {
        return 0.0;
    }
    if (value < BSPLINE_EPSILON) {
        return 0.0;
    }
    return value;
}

// pochodna funkcji interpolujacej
__device__ Precision Nprim(const Precision x) {

    Precision x_abs = fabs(x);
    if (x_abs < 1.0) {
        return (1.5 * x * x_abs - 2.0 * x);
    }
    else if (x_abs < 2.0) {
        return (-0.5 * x * x_abs + 2.0 * x - 2.0 * x / x_abs);
    }
    else {
        return 0.0;
    }
}

// zwraca indeks wezla siatki w tablicy wezlow, na podstawie jego wspolrzednych podzielonych przez h
__host__ __device__ int grid_index(int x, int y, int z) {
    return n_z_grid * n_y_grid * x + n_z_grid * y + z;
}

// funkcja do wyznaczania wag interpolacji
__device__ Precision w(const Vector xp, const Vector xg) {
    return N(1.0 / h * (xp.x() - xg.x())) * N(1.0 / h * (xp.y() - xg.y())) * N(1.0 / h * (xp.z() - xg.z()));
}

// gradient funkcji wagowej
__device__ Vector gradw(const Vector xp, const Vector xg) {
    return 1.0 / h * Vector(Nprim(1.0 / h * (xp.x() - xg.x())) * N(1.0 / h * (xp.y() - xg.y())) * N(1.0 / h * (xp.z() - xg.z())), N(1.0 / h * (xp.x() - xg.x())) * Nprim(1.0 / h * (xp.y() - xg.y())) * N(1.0 / h * (xp.z() - xg.z())), N(1.0 / h * (xp.x() - xg.x())) * N(1.0 / h * (xp.y() - xg.y())) * Nprim(1.0 / h * (xp.z() - xg.z())));
}

// odwracanie macierzy, brak sprawdzania, czy nie jest osobliwa
// ale jesli by byla, to wychwycenie tego i tak nic nie da
__device__ Matrix invert(const Matrix& mat) {

    Matrix cofactorMat;

    cofactorMat(0, 0) = mat(1, 1) * mat(2, 2) - mat(1, 2) * mat(2, 1);
    cofactorMat(0, 1) = -(mat(1, 0) * mat(2, 2) - mat(1, 2) * mat(2, 0));
    cofactorMat(0, 2) = mat(1, 0) * mat(2, 1) - mat(1, 1) * mat(2, 0);

    cofactorMat(1, 0) = -(mat(0, 1) * mat(2, 2) - mat(0, 2) * mat(2, 1));
    cofactorMat(1, 1) = mat(0, 0) * mat(2, 2) - mat(0, 2) * mat(2, 0);
    cofactorMat(1, 2) = -(mat(0, 0) * mat(2, 1) - mat(0, 1) * mat(2, 0));

    cofactorMat(2, 0) = mat(0, 1) * mat(1, 2) - mat(0, 2) * mat(1, 1);
    cofactorMat(2, 1) = -(mat(0, 0) * mat(1, 2) - mat(0, 2) * mat(1, 0));
    cofactorMat(2, 2) = mat(0, 0) * mat(1, 1) - mat(0, 1) * mat(1, 0);

    Precision determinant = mat(0, 0) * cofactorMat(0, 0) +
        mat(0, 1) * cofactorMat(0, 1) +
        mat(0, 2) * cofactorMat(0, 2);

    return (1.0 / determinant) * cofactorMat.transpose();
}

// svd znalezione w internecie, w svd3_cuda.h jest zrodlo
__device__ void computeSVD(const Matrix& a, Matrix& u, Matrix& s, Matrix& v) {
    svd(a(0), a(3), a(6), a(1), a(4), a(7), a(2), a(5), a(8),
        u(0), u(3), u(6), u(1), u(4), u(7), u(2), u(5), u(8),
        s(0), s(3), s(6), s(1), s(4), s(7), s(2), s(5), s(8),
        v(0), v(3), v(6), v(1), v(4), v(7), v(2), v(5), v(8));
}

// tensor naprezen Cauchy'ego
__device__ Matrix sigma(const Matrix Fe, const Matrix Fp, MaterialProperties& materialProps) {

    Matrix U;
    Matrix V;
    Matrix S;

    computeSVD(Fe, U, S, V);

    Matrix Re = U * V.transpose();

    Precision Jp = Fp.determinant();
    Precision Je = Fe.determinant();

    Precision mu = materialProps.mu_0 * exp(materialProps.xi * (1.0 - Jp));
    Precision lambda = materialProps.lambda_0 * exp(materialProps.xi * (1.0 - Jp));

    Matrix sigma = 2.0 * mu / Jp * (Fe - Re) * Fe.transpose() + lambda / Jp * (Je - 1.0) * Je * Matrix::Identity();

    return sigma;
}


// resetowanie wartosci na siatce
__global__ void reset_grid(Node* Grid, const size_t grid_size) {

    // zrownoleglanie po wezlach siatki

    int node_idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (node_idx >= grid_size) return;

    Node& node = Grid[node_idx];

    node.m = 0.0f;
    node.v_old = Vector::Zero();
    node.f = Vector::Zero();
}

// interpolacja mas czastek na siatke
// uzywana jedynie na poczatku, zeby wyznaczyc objetosc czastek
__global__ void particle_to_grid_mass(Node* Grid, Particle* Particles, const size_t grid_size, const size_t particles_count) {

    // zrownoleglanie po czastkach

    int particle_idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (particle_idx >= particles_count) return;

    Particle& p = Particles[particle_idx];

    // indeks wezla siatki znajdujacego sie najblizej czastki
    int px = static_cast<int>((p.x.x() - x_min_grid) / h);
    int py = static_cast<int>((p.x.y() - y_min_grid) / h);
    int pz = static_cast<int>((p.x.z() - z_min_grid) / h);

// idziemy tylko po wezlach siatki znajdujacych sie najblizej czastki (64 wezly)
    for (int z = -1; z < 3; z++) {
        for (int y = -1; y < 3; y++) {
            for (int x = -1; x < 3; x++) {

                if (px + x > 0 && px + x < n_x_grid && py + y > 0 && py + y < n_y_grid && pz + z > 0 && pz + z < n_z_grid) {

                    int node_idx = grid_index(px + x, py + y, pz + z);
                    Precision weight = w(p.x, Grid[node_idx].x);
                    atomicAdd(&Grid[node_idx].m, weight * p.m);
                    p.weights[x + 1][y + 1][z + 1] = weight;

                }
            }
        }
    }
}

// wyznaczanie objetosci czastek na poczatku obliczen
__global__ void calculate_V(const Node* Grid, Particle* Particles, const size_t grid_size, const size_t particles_count) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count) return;

    Particle& particle = Particles[pIdx];

    // indeks wezla siatki znajdujacego sie najblizej czastki
    int px = static_cast<int>((particle.x.x() - x_min_grid) / h);
    int py = static_cast<int>((particle.x.y() - y_min_grid) / h);
    int pz = static_cast<int>((particle.x.z() - z_min_grid) / h);

    Precision rhop = 0;

    for (int z = -1; z < 3; z++) {
        for (int y = -1; y < 3; y++) {
            for (int x = -1; x < 3; x++) {

                if (px + x > 0 && px + x < n_x_grid && py + y > 0 && py + y < n_y_grid && pz + z < n_z_grid) {
                    int node_idx = grid_index(px + x, py + y, pz + z);
                    Precision weight = particle.weights[x + 1][y + 1][z + 1];
                    if (weight > BSPLINE_EPSILON) {
                        rhop += Grid[node_idx].m * weight / (h * h * h);
                    }
                }
            }
        }
    }
    particle.V = Particles[pIdx].m / rhop;
}

// interpolacja mas i predkosci z czastek na siatke
// prawdopodobnie zajmuje najwiecej czasu ze wzgledu na uzywanie atomicAdd
// inne rozwiazania bylyby bardziej skomplikowane do zaimplementowania

// poza interpolacja zapamietywane sa tez wagi i gradienty wag interpolacji,
// dzieki czemu nie trzeba ich liczyc ponownie przy interpolacji z siatki na czastki

// ze wzgledu na funkcje interpolacyjna, czastka moze wplywac tylko na 2 najblizsze wezly siatki w kazda strone
// zatem nie trzeba liczyc wszystkich, wystarczy tylko najblizsze 64

// dla zaoszczedzenia wywolan, juz w tej funkcji liczone sa sily w wezlach

__global__ void particle_to_grid(Node* Grid, Particle* Particles, const size_t grid_size, const size_t particles_count, MaterialProperties& materialProps) {

    // zrownoleglanie po czastkach

    int particle_idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (particle_idx >= particles_count) return;

    Particle& p = Particles[particle_idx];

    // indeks wezla siatki znajdujacego sie najblizej czastki
    int px = static_cast<int>((p.x.x() - x_min_grid) / h);
    int py = static_cast<int>((p.x.y() - y_min_grid) / h);
    int pz = static_cast<int>((p.x.z() - z_min_grid) / h);

    for (int z = -1; z < 3; z++) {
        for (int y = -1; y < 3; y++) {
            for (int x = -1; x < 3; x++) {

                if (px + x > 0 && px + x < n_x_grid && py + y > 0 && py + y < n_y_grid && pz + z > 0 && pz + z < n_z_grid) {

                    int node_idx = grid_index(px + x, py + y, pz + z);
                    Precision weight = w(p.x, Grid[node_idx].x);
                    Vector weight_gradient = gradw(p.x, Grid[node_idx].x);

                    atomicAdd(&Grid[node_idx].m, weight * p.m);

                    atomicAdd(&Grid[node_idx].v_old[0], weight * p.m * p.v.x());
                    atomicAdd(&Grid[node_idx].v_old[1], weight * p.m * p.v.y());
                    atomicAdd(&Grid[node_idx].v_old[2], weight * p.m * p.v.z());

                    p.weights[x + 1][y + 1][z + 1] = weight;
                    p.weight_gradients[x + 1][y + 1][z + 1] = weight_gradient;


                    if (weight > BSPLINE_EPSILON) {
                        Matrix FeFp = p.Fe * p.Fp;

                        Precision determinant = FeFp.determinant();
                        Matrix sigma_value = sigma(p.Fe, p.Fp, materialProps);
                        Vector f_to_add = -p.V * determinant * sigma_value * weight_gradient;
                        atomicAdd(&Grid[node_idx].f[0], f_to_add[0]);
                        atomicAdd(&Grid[node_idx].f[1], f_to_add[1]);
                        atomicAdd(&Grid[node_idx].f[2], f_to_add[2]);
                    }

                }
            }
        }
    }
}


// sprawdzamy, czy cokolwiek zostalo zinterpolowane do wezla, jezeli nie, to tak naprawde w dalszej czesci kroku obliczeniowego nie powinnismy sie nim zajmowac

__global__ void node_not_zero(Node* Grid, const size_t grid_size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= grid_size) return;

    Node& node = Grid[idx];
    if (node.m == 0.0f) {
        node.v_old = Vector::Zero();
    }
    else {
        node.v_old = node.v_old / node.m;
    }
}


// aktualizacja predkosci w wezlach
__global__ void grid_velocity(Node* Grid, const size_t n_grid_nodes, const Precision dt) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n_grid_nodes) {

        if (Grid[idx].m != 0.0f) {
            //Grid[idx].v += dt * (Grid[idx].f / Grid[idx].m);
            Grid[idx].v = Grid[idx].v_old + dt * (Grid[idx].f / Grid[idx].m + Vector(0.0, 0.0, -9.81));
        }
    }
}

// powinna byc w pliku z SDF, ale wtedy nie kompiluje sie
__host__ __device__ int grid_index_sdf(int x, int y, int z) {
    return n_z_grid * n_y_grid * x + n_z_grid * y + z;
}

// na podstawie sdf_grid zwraca odleglosc ze znakiem od danego punktu point
__host__ __device__ Precision signed_distance_interpolation(const Vector& point, SDF_Node* sdf_grid) {

    // trzeba bedzie zmienic jak beda wczytywane dowolne modele (albo normalizowac wspolrzedne wczytywanych modeli na, zeby zaczynaly sie od 0
    // ^ chyba dziala, ale nie dam glowy
    int px = static_cast<int>((point.x() - x_min_grid) / h);
    int py = static_cast<int>((point.y() - y_min_grid) / h);
    int pz = static_cast<int>((point.z() - z_min_grid) / h);

    int index[8] = {
        grid_index_sdf(px, py, pz),
        grid_index_sdf(px + 1, py, pz),
        grid_index_sdf(px, py + 1, pz),
        grid_index_sdf(px, py, pz + 1),
        grid_index_sdf(px + 1, py + 1, pz),
        grid_index_sdf(px, py + 1, pz + 1),
        grid_index_sdf(px + 1, py, pz + 1),
        grid_index_sdf(px + 1, py + 1, pz + 1)
    };

    Precision numerator = 0.0f;
    Precision denominator = 0.0f;

    for (int i = 0; i < 8; i++) {
        Vector diff = point - sdf_grid[index[i]].x;
        Precision dist_squared = diff.transpose() * diff;
        if (dist_squared < BSPLINE_EPSILON) {
            return sdf_grid[index[i]].signed_dist;
        }

        Precision weight = 1.0f / dist_squared;
        numerator += weight * sdf_grid[index[i]].signed_dist;
        denominator += weight;
    }

    return numerator / denominator;
};

// na podstawie sdf_grid zwraca wektor wodzacy od punktu point do sciany
__device__ Vector normal_vector_interpolation(const Vector& point, SDF_Node* sdf_grid) {

    // trzeba bedzie zmienic jak beda wczytywane dowolne modele (albo normalizowac wspolrzedne wczytywanych modeli na, zeby zaczynaly sie od 0
    // ^ chyba dziala, ale nie dam glowy
    int px = static_cast<int>((point.x() - x_min_grid) / h);
    int py = static_cast<int>((point.y() - y_min_grid) / h);
    int pz = static_cast<int>((point.z() - z_min_grid) / h);

    int index[8] = {
        grid_index_sdf(px, py, pz),
        grid_index_sdf(px + 1, py, pz),
        grid_index_sdf(px, py + 1, pz),
        grid_index_sdf(px, py, pz + 1),
        grid_index_sdf(px + 1, py + 1, pz),
        grid_index_sdf(px, py + 1, pz + 1),
        grid_index_sdf(px + 1, py, pz + 1),
        grid_index_sdf(px + 1, py + 1, pz + 1)
    };

    Vector numerator = Vector::Zero();
    Precision denominator = 0.0f;

    for (int i = 0; i < 8; i++) {
        Vector diff = point - sdf_grid[index[i]].x;
        Precision dist_squared = diff.transpose() * diff;
        if (dist_squared < BSPLINE_EPSILON) {  // If the point is exactly at one of the known points
            return sdf_grid[index[i]].dist_vec;
        }

        Precision weight = 1.0f / dist_squared;
        numerator += weight * sdf_grid[index[i]].dist_vec;
        denominator += weight;
    }

    return numerator / denominator;
};


// obliczenia kolizji na siatce
// polega to na sprawdzeniu, czy po hipotetycznym przesunieciu wezla siatki nie wyjdzie on poza domene obliczen
// jesli by weszedl to trzeba zmienic predkosc w nim: wyzerowac rzut wektora predkosci na wektor normalny do sciany, a wektor styczny przemnozyc przez wspolczynnik "przylepiania" do sciany
__global__ void grid_collisions(Node* Grid, SDF_Node* sdf_gpu, const size_t n_grid_nodes, const Precision dt) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n_grid_nodes) {

        if (Grid[idx].m != 0.0f) {

            Vector new_pos = Grid[idx].x + Grid[idx].v * dt;

            Precision signed_distance = signed_distance_interpolation(new_pos, sdf_gpu);

            if (signed_distance < 0.0) {

                Vector normal = normal_vector_interpolation(Grid[idx].x, sdf_gpu);// .normalized();

                
                Vector proj_v_on_normal = (Grid[idx].v.dot(normal) / normal.dot(normal)) * normal;
                Grid[idx].v = Grid[idx].v - proj_v_on_normal;
                Grid[idx].v *= device_sticky;

                //Grid[idx].v = Vector::Zero();
            }
        }
    }
}


// uaktualnienie tensorow deformacji,
// wymaga wyznaczenia SVD
__global__ void update_deformation_gradient(const Node* Grid, Particle* Particles, const size_t n_grid_nodes, const size_t particles_count, const Precision dt, MaterialProperties& materialProps) {
    int p = blockIdx.x * blockDim.x + threadIdx.x;

    if (p < particles_count) {
        Particle& particle = Particles[p];

        Matrix grad_v = particle.velocity_gradient;

        Matrix Fe_part = (Matrix::Identity() + dt * grad_v) * particle.Fe;

        Matrix F_p = Fe_part * particle.Fp;

        Matrix U;
        Matrix V;
        Matrix Sigma;

        computeSVD(Fe_part, U, Sigma, V);

        if (Sigma(0, 0) < 1 - materialProps.theta_c) {
            Sigma(0, 0) = 1 - materialProps.theta_c;
        }
        else if (Sigma(0, 0) > 1.0 + materialProps.theta_s) {
            Sigma(0, 0) = 1.0 + materialProps.theta_s;
        }

        if (Sigma(1, 1) < 1 - materialProps.theta_c) {
            Sigma(1, 1) = 1 - materialProps.theta_c;
        }
        else if (Sigma(1, 1) > 1.0 + materialProps.theta_s) {
            Sigma(1, 1) = 1.0 + materialProps.theta_s;
        }

        if (Sigma(2, 2) < 1 - materialProps.theta_c) {
            Sigma(2, 2) = 1 - materialProps.theta_c;
        }
        else if (Sigma(2, 2) > 1.0 + materialProps.theta_s) {
            Sigma(2, 2) = 1.0 + materialProps.theta_s;
        }

        particle.Fp = V * invert(Sigma) * U.transpose() * F_p;
        particle.Fe = U * Sigma * V.transpose();
        
    }
}

// uaktualnienie predkosci czastek
// idziemy po czastkach, 
// dla kazdej czastki przeprowadzamy interpolacje z odpowiadajacych jej wezlow na czastke,
// nowa predkosc czastki wyznaczamy z polaczaniea schmatow PIC i FLIP
// na tym etapie wyznaczamy tez gradient predkosci do nastepnego kroku czasowego
__global__ void update_particle_velocities(const Node* Grid, Particle* Particles, const size_t grid_size, const size_t particles_count, SimulationProperties& simProps) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count) return;

    Particle& particle = Particles[pIdx];
    Vector v_flip = particle.v;
    Vector v_pic = Vector::Zero();
    particle.velocity_gradient = Matrix::Zero();


    // indeks wezla siatki znajdujacego sie najblizej czastki
    int px = static_cast<int>((particle.x.x() - x_min_grid) / h);
    int py = static_cast<int>((particle.x.y() - y_min_grid) / h);
    int pz = static_cast<int>((particle.x.z() - z_min_grid) / h);

    for (int z = -1; z < 3; z++) {
        for (int y = -1; y < 3; y++) {
            for (int x = -1; x < 3; x++) {

                if (px + x > 0 && px + x < n_x_grid && py + y > 0 && py + y < n_y_grid && pz + z < n_z_grid) {

                    Precision weight = particle.weights[x + 1][y + 1][z + 1];

                    if (weight > BSPLINE_EPSILON) {
                        int node_idx = grid_index(px + x, py + y, pz + z);
                        Vector gradw_value = particle.weight_gradients[x + 1][y + 1][z + 1];
                        v_pic += Grid[node_idx].v * weight;
                        v_flip += (Grid[node_idx].v - Grid[node_idx].v_old) * weight;
                        Matrix to_add = Grid[node_idx].v * gradw_value.transpose();
                        particle.velocity_gradient = particle.velocity_gradient + to_add;
                    }
                }
            }
        }
    }

    particle.v = (1.0f - simProps.alpha) * v_pic + simProps.alpha * v_flip;
}


// przeliczenie kolizji na czastkach
// na wypadek, jakby pozostaly jakies bledy po przeliczeniu na siatce
// sprawdzamy, czy hipotetycznie przy danej predkosci czastka znalazlaby sie poza domena obliczen
// jesli tak, to zmieniamy wektor predkosci, usuwajac skladowa normalna
// dodatkowo powodujac pewne odbicie od scianki, proporcjonalne do dotychczasowego wektora normalnego i wspolczynnika "sticky"
__global__ void particle_collisions(Particle* Particles, SDF_Node* sdf_gpu, const size_t particles_count, const Precision dt) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count) return;

    Vector new_pos = Particles[pIdx].x + Particles[pIdx].v * dt;

    Precision signed_distance = signed_distance_interpolation(new_pos, sdf_gpu);

    if (signed_distance < 0.0) {

        Vector normal = normal_vector_interpolation(Particles[pIdx].x, sdf_gpu);// .normalized();
        Vector proj_v_on_normal = (Particles[pIdx].v.dot(normal) / normal.dot(normal)) * normal;

        Particles[pIdx].v = Particles[pIdx].v - (1 + device_sticky) * proj_v_on_normal;
    }
}


// aktualizacja polozen czastek
// schemat Eulera, jawny pierwszego rzedu dokladnosci w czasie
__global__ void update_particle_positions(Particle* Particles, const size_t particles_count, const Precision dt) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count) return;

    Particles[pIdx].x += Particles[pIdx].v * dt;
}


// EMITERY nie zostaly przetestowane
// prawdopodobnie wymagaja rozwiazania kilku problemow

// aktualizacja emiterow, czyli dodawanie nowych czastek 
// wstawia czastke w pamiec
__global__ void update_emiters(Particle* Particles, const size_t particles_count, const size_t particles_count_new, const Emitter* Emitters, const size_t emitters_count, const Eigen::Vector3i* triangles_vertex, const Eigen::Vector3i* triangles_normal, const size_t triangles_count, const Vector* vertices, const size_t vertices_count, const Vector* normals, size_t seed, MaterialProperties& materialProps, SimulationProperties& simProps) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count_new) return;


    // wybrac emiter z listy
    for (int i = 0; i < emitters_count; i++) {
        if ((pIdx >= Emitters[i].starting_particle_index) && (pIdx < Emitters[i].ending_particle_index)) {
            const Emitter& emitter = Emitters[i];

            pIdx = pIdx + particles_count; // to moze powinno byc po curand_init

            // dodac na jego powierzchni czastki z predkoscia normalna do powierzchni
            // czyli dla emitera wybieram trojkat losowo

            curandState state;
            curand_init(seed, pIdx, 0, &state);
            
            int randomIndex = emitter.starting_triangle_index + curand(&state) % emitter.n_triangles;

            // pobranie wierzcholkow
            Eigen::Vector3i tri = triangles_vertex[randomIndex];
            Vector v0 = vertices[tri[0]];
            Vector v1 = vertices[tri[1]];
            Vector v2 = vertices[tri[2]];

            // nastepnie punkt z trojkata losowo, wspolrzedne barycentryczne
            Precision u = curand_uniform(&state);
            Precision v = curand_uniform(&state);

            if (u + v > 1.0f) {
                u = 1.0f - u;
                v = 1.0f - v;
            }

            Vector initial_position = (1 - u - v) * v0 + u * v1 + v * v2;


            // wektor normalny i przemnozenie go przez wartosc predkosci, zeby uzyskac wektor predkosci poczatkowej
            Vector normal = normals[randomIndex];
            Vector initial_velocity = normal * emitter.particle_initial_velocity;
            
            // masa jest zadawana na podstawie gestosci i objetosci wyznaczonych ogolnie dla czastek, wszystkie maja taka sama mase
            Particles[pIdx] = Particle(initial_position, initial_velocity, materialProps.rho_0, simProps.particle_volume);

        }
    }
}

// liczy objetosc nowych czastek
__global__ void emitter_calculate_V(const Node* Grid, Particle* Particles, const size_t grid_size, const size_t particles_count, const size_t particles_count_new) {
    int pIdx = blockIdx.x * blockDim.x + threadIdx.x;
    if (pIdx >= particles_count_new) return;

    Particle& particle = Particles[pIdx + particles_count];

    // indeks wezla siatki znajdujacego sie najblizej czastki
    int px = static_cast<int>((particle.x.x() - x_min_grid) / h);
    int py = static_cast<int>((particle.x.y() - y_min_grid) / h);
    int pz = static_cast<int>((particle.x.z() - z_min_grid) / h);

    Precision rhop = 0;

    for (int z = -1; z < 3; z++) {
        for (int y = -1; y < 3; y++) {
            for (int x = -1; x < 3; x++) {

                if (px + x > 0 && px + x < n_x_grid && py + y > 0 && py + y < n_y_grid && pz + z < n_z_grid) {
                    int node_idx = grid_index(px + x, py + y, pz + z);
                    Precision weight = particle.weights[x + 1][y + 1][z + 1];
                    if (weight > BSPLINE_EPSILON) {
                        rhop += Grid[node_idx].m * weight / (h * h * h);
                    }
                }
            }
        }
    }
    particle.V = Particles[pIdx].m / rhop;
}


// krok w metodzie mpm
__host__ void MPM_step(Particle* particles_gpu, Node* grid_gpu, SDF_Node* sdf_gpu, size_t& n_particles, size_t n_grid_nodes, Precision dt, int blocksPerGrid, int blocksPerParticles, int threadsPerBlock, Emitter* emitters_gpu, size_t n_emitters, Eigen::Vector3i* emitters_triangles_vertex_gpu, Eigen::Vector3i* emitters_triangles_normal_gpu, size_t n_emitter_triangles, Vector* emitter_vertices, size_t n_emitter_vertices, Vector* emitter_normals, size_t seed, size_t n_new_particles, MaterialProperties* materialProps_gpu, SimulationProperties* simProps_gpu) {
    // zerowanie wartości na siatce (0) (GPU)
    
    reset_grid << <blocksPerGrid, threadsPerBlock >> > (grid_gpu, n_grid_nodes);
    //cudaDeviceSynchronize();
    

    // dodanie czastek w emiterach
    update_emiters << <blocksPerParticles, threadsPerBlock >> > (particles_gpu, n_particles, n_new_particles, emitters_gpu, n_emitters, emitters_triangles_vertex_gpu, emitters_triangles_normal_gpu, n_emitter_triangles, emitter_vertices, n_emitter_vertices, emitter_normals, seed, *materialProps_gpu, *simProps_gpu);

    
    // interpolacja z czastek na siatke (1) (GPU) wyznaczenie sil na siatce (3)
    particle_to_grid << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles + n_new_particles, *materialProps_gpu);
    //cudaDeviceSynchronize();
    

    //particle_to_grid_velocity<<<blocksPerGrid, threadsPerBlock>>>(grid_gpu, particles_gpu, n_grid_nodes, Particles.size());
    //cudaDeviceSynchronize();

    // poprawki na siatce (unikniecie dzielenia przez 0, gdy masa w wezle jest rowna 0)
    // w przyszlosci nie powinno byc potrzebne, w p2g powinnismy ustawiac wezlom flage, ze sa aktywne
    node_not_zero << <blocksPerGrid, threadsPerBlock >> > (grid_gpu, n_grid_nodes);
    //cudaDeviceSynchronize();


    // polczenie objetosci nowch czastek
    emitter_calculate_V << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles, n_new_particles);
    
    n_particles += n_new_particles;

    // wyznaczenie sil na siatce (3) GPU, przeniesione do punktu (1)
    //grid_forces<<<blocksPerGrid, threadsPerBlock>>>(grid_gpu, particles_gpu, n_grid_nodes, Particles.size());
    //cudaDeviceSynchronize();

    // wyznaczenie predkosci na siatce(4) GPU
    // wykonywac tylko w aktywnych wezlach
    // w sumie mozna by sie zastanowic, czy nie tworzyc na poczatku tablicy wskaznikow na aktywne wezly w danym kroku, zamiast aktywowac
    // wtedy mniej watkow byloby potrzebne
    grid_velocity << <blocksPerGrid, threadsPerBlock >> > (grid_gpu, n_grid_nodes, dt);
    //cudaDeviceSynchronize();

    // kolizje na siatce (5) GPU
    grid_collisions << <blocksPerGrid, threadsPerBlock >> > (grid_gpu, sdf_gpu, n_grid_nodes, dt);
    //cudaDeviceSynchronize();

    // aktualizacja gradientow odksztalcen czastek (7) GPU
    update_deformation_gradient << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles, dt, *materialProps_gpu);
    //cudaDeviceSynchronize();


    // aktualizacja predkosci czastek (8) GPU
    update_particle_velocities << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles, *simProps_gpu);
    //cudaDeviceSynchronize();

    // kolizje cząstki (9) GPU
    particle_collisions << <blocksPerParticles, threadsPerBlock >> > (particles_gpu, sdf_gpu, n_particles, dt);
    //cudaDeviceSynchronize();

    // wyznaczenie nowych polozen (10) GPU
    update_particle_positions << <blocksPerParticles, threadsPerBlock >> > (particles_gpu, n_particles, dt);
    cudaDeviceSynchronize();
    

}

struct XYZ {
    double x, y, z;
};

struct TRIANGLE {
    XYZ p[3];
};

struct GRIDCELL {
    XYZ p[8];   // chyba nieefektywne pamieciowo, chyba powinien przechowywac wskazniki na swoje wierzcholki
                // jakby przechowywal wskazniki, to dodatkowo w siatce nie powtarzaly by sie wierzcholki
                // ale podejrzewam, ze wtedy mogloby sie to mniej przyjemnie kopiowac na gpu
    double val[8];
};


// Linearly interpolate the position where an isosurface cuts an edge between two vertices
__device__ XYZ vertexInterp(float isolevel, XYZ p1, XYZ p2, double valp1, double valp2) {
    double mu;
    XYZ p;

    if (fabs(isolevel - valp1) < device_MCE)
        return p1;
    if (fabs(isolevel - valp2) < device_MCE)
        return p2;
    if (fabs(valp1 - valp2) < device_MCE)
        return p1;

    mu = (isolevel - valp1) / (valp2 - valp1);
    p.x = p1.x + mu * (p2.x - p1.x);
    p.y = p1.y + mu * (p2.y - p1.y);
    p.z = p1.z + mu * (p2.z - p1.z);

    return p;
}

__constant__ int edgeTable[256] = {
    0x0 , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
    0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
    0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
    0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
    0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
    0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
    0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
    0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
    0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
    0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
    0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
    0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
    0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
    0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
    0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
    0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
    0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
    0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
    0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
    0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
    0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
    0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
    0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
    0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
    0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
    0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
    0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
    0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
    0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
    0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
    0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
    0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0
};
__constant__ int triTable[256][16] = {
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
    {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
    {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
    {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
    {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
    {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
    {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
    {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
    {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
    {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
    {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
    {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
    {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
    {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
    {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
    {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
    {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
    {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
    {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
    {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
    {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
    {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
    {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
    {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
    {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
    {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
    {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
    {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
    {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
    {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
    {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
    {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
    {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
    {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
    {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
    {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
    {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
    {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
    {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
    {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
    {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
    {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
    {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
    {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
    {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
    {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
    {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
    {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
    {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
    {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
    {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
    {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
    {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
    {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
    {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
    {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
    {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
    {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
    {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
    {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
    {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
    {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
    {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
    {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
    {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
    {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
    {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
    {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
    {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
    {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
    {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
    {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
    {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
    {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
    {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
    {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
    {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
    {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
    {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
    {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
    {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
    {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
    {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
    {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
    {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
    {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
    {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
    {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
    {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
    {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
    {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
    {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
    {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
    {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
    {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
    {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
    {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};


__global__ void polygoniseKernel(GRIDCELL* dev_grids, float isolevel, TRIANGLE* dev_triangles, int numCells, int* triangleCounter) {

    int idx = blockIdx.x * blockDim.x + threadIdx.x; // index of current grid cell
    if (idx >= numCells) return;

    int cubeindex = 0;
    XYZ vertlist[12];
    GRIDCELL grid = dev_grids[idx];


    // Determine the index into the edge table
    if (grid.val[0] < isolevel) cubeindex |= 1;
    if (grid.val[1] < isolevel) cubeindex |= 2;
    if (grid.val[2] < isolevel) cubeindex |= 4;
    if (grid.val[3] < isolevel) cubeindex |= 8;
    if (grid.val[4] < isolevel) cubeindex |= 16;
    if (grid.val[5] < isolevel) cubeindex |= 32;
    if (grid.val[6] < isolevel) cubeindex |= 64;
    if (grid.val[7] < isolevel) cubeindex |= 128;


    // Cube is entirely in/out of the surface
    if (edgeTable[cubeindex] == 0)
        return;

    // Find the vertices where the surface intersects the cube
    if (edgeTable[cubeindex] & 1)
        vertlist[0] = vertexInterp(isolevel, grid.p[0], grid.p[1], grid.val[0], grid.val[1]);
    if (edgeTable[cubeindex] & 2)
        vertlist[1] = vertexInterp(isolevel, grid.p[1], grid.p[2], grid.val[1], grid.val[2]);
    if (edgeTable[cubeindex] & 4)
        vertlist[2] = vertexInterp(isolevel, grid.p[2], grid.p[3], grid.val[2], grid.val[3]);
    if (edgeTable[cubeindex] & 8)
        vertlist[3] = vertexInterp(isolevel, grid.p[3], grid.p[0], grid.val[3], grid.val[0]);
    if (edgeTable[cubeindex] & 16)
        vertlist[4] = vertexInterp(isolevel, grid.p[4], grid.p[5], grid.val[4], grid.val[5]);
    if (edgeTable[cubeindex] & 32)
        vertlist[5] = vertexInterp(isolevel, grid.p[5], grid.p[6], grid.val[5], grid.val[6]);
    if (edgeTable[cubeindex] & 64)
        vertlist[6] = vertexInterp(isolevel, grid.p[6], grid.p[7], grid.val[6], grid.val[7]);
    if (edgeTable[cubeindex] & 128)
        vertlist[7] = vertexInterp(isolevel, grid.p[7], grid.p[4], grid.val[7], grid.val[4]);
    if (edgeTable[cubeindex] & 256)
        vertlist[8] = vertexInterp(isolevel, grid.p[0], grid.p[4], grid.val[0], grid.val[4]);
    if (edgeTable[cubeindex] & 512)
        vertlist[9] = vertexInterp(isolevel, grid.p[1], grid.p[5], grid.val[1], grid.val[5]);
    if (edgeTable[cubeindex] & 1024)
        vertlist[10] = vertexInterp(isolevel, grid.p[2], grid.p[6], grid.val[2], grid.val[6]);
    if (edgeTable[cubeindex] & 2048)
        vertlist[11] = vertexInterp(isolevel, grid.p[3], grid.p[7], grid.val[3], grid.val[7]);


    // How many triangles are generated for current grid cell
    int ntriang = 0;
    for (int i = 0; triTable[cubeindex][i] != -1; i += 3) {
        ntriang++;
    }

    // Get the current triangle index and update the counter
    int triangleIndex = atomicAdd(triangleCounter, ntriang);

    // Create triangles
    for (int i = 0; triTable[cubeindex][i] != -1; i += 3) {
        dev_triangles[triangleIndex].p[0] = vertlist[triTable[cubeindex][i]];
        dev_triangles[triangleIndex].p[1] = vertlist[triTable[cubeindex][i + 1]];
        dev_triangles[triangleIndex].p[2] = vertlist[triTable[cubeindex][i + 2]];
        triangleIndex++;
    }

}

cudaError_t polygoniseWithCuda(GRIDCELL* grids, TRIANGLE* triangles, int numCells, float isolevel, int* totalTriangles) {

    GRIDCELL* dev_grids = nullptr;
    TRIANGLE* dev_triangles = nullptr;
    int* triangleCounter = nullptr;

    cudaError_t cudaStatus;


    // Choose which GPU to run on, change this on a multi-GPU system.
    cudaStatus = cudaSetDevice(0);
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaSetDevice failed!  Do you have a CUDA-capable GPU installed?");
        goto Error;
    }

    // Allocate GPU buffers for grids

    // ten blad ponizej moze oznaczac wiele rzeczy
    // np. ze wzgledu na to, ze wczesniej nie ma kontroli, to moze oznaczac, ze w obliczeniach wystapil nan ze wzgedu na zbyt duzy krok czasowy
    cudaStatus = cudaMalloc((void**)&dev_grids, numCells * sizeof(GRIDCELL));
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMalloc for dev_grids failed!");
        goto Error;
    }


    // Allocate GPU buffers for triangles (max 5 triangles per cell)
    cudaStatus = cudaMalloc((void**)&dev_triangles, numCells * 5 * sizeof(TRIANGLE));
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMalloc for dev_triangles failed!");
        goto Error;
    }

    // Allocate GPU buffers for triangleCounter
    cudaStatus = cudaMalloc((void**)&triangleCounter, sizeof(int));
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMalloc for triangleCounter failed!");
        goto Error;
    }

    // Copy grid data from host memory to GPU buffers
    cudaStatus = cudaMemcpy(dev_grids, grids, numCells * sizeof(GRIDCELL), cudaMemcpyHostToDevice);
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMemcpy for grids failed!");
        goto Error;
    }

    // to nie bylo wczesniej potrzebne
    cudaStatus = cudaMemset(triangleCounter, 0, sizeof(int));


    // Launch a kernel on the GPU
    int threadsPerBlock = 256;
    int blocksPerGrid = (numCells + threadsPerBlock - 1) / threadsPerBlock;
    polygoniseKernel << <blocksPerGrid, threadsPerBlock >> > (dev_grids, isolevel, dev_triangles, numCells, triangleCounter);

    // Check for any errors launching the kernel
    cudaStatus = cudaGetLastError();
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "PolygoniseKernel launch failed: %s\n", cudaGetErrorString(cudaStatus));
        goto Error;
    }

    // cudaDeviceSynchronize waits for the kernel to finish, and returns
    // any errors encountered during the launch.
    cudaStatus = cudaDeviceSynchronize();
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaDeviceSynchronize returned error code %d after launching PolygoniseKernel!\n", cudaStatus);
        goto Error;
    }

    // Copy triangleCounter from GPU buffer to host memory.
    cudaStatus = cudaMemcpy(totalTriangles, triangleCounter, sizeof(int), cudaMemcpyDeviceToHost);
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMemcpy for totalTriangles failed!");
        goto Error;
    }

    // Copy triangles from GPU buffer to host memory
    cudaStatus = cudaMemcpy(triangles, dev_triangles, numCells * 5 * sizeof(TRIANGLE), cudaMemcpyDeviceToHost);
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMemcpy for triangles failed!");
        goto Error;
    }

Error:
    cudaFree(dev_grids);
    cudaFree(dev_triangles);
    cudaFree(triangleCounter);

    return cudaStatus;
}


void update_MC_Grid_Cell_Values(std::vector<Node>& Grid, std::vector<GRIDCELL>& gridCells, int MC_n_x_grid, int MC_n_y_grid, int MC_n_z_grid) {
    int cellIndex = 0; // indeks komórki w gridCells

    for (int i = 0; i < MC_n_x_grid; i++) {
        for (int j = 0; j < MC_n_y_grid; j++) {
            for (int k = 0; k < MC_n_z_grid; k++) {
                // indeksy

                int idx000 = i * n_y_grid * n_z_grid + j * n_z_grid + k;
                int idx001 = (i + 1) * n_y_grid * n_z_grid + j * n_z_grid + k;
                int idx010 = (i + 1) * n_y_grid * n_z_grid + (j + 1) * n_z_grid + k;
                int idx011 = i * n_y_grid * n_z_grid + (j + 1) * n_z_grid + k;
                int idx100 = i * n_y_grid * n_z_grid + j * n_z_grid + (k + 1);
                int idx101 = (i + 1) * n_y_grid * n_z_grid + j * n_z_grid + (k + 1);
                int idx110 = (i + 1) * n_y_grid * n_z_grid + (j + 1) * n_z_grid + (k + 1);
                int idx111 = i * n_y_grid * n_z_grid + (j + 1) * n_z_grid + (k + 1);


                gridCells[cellIndex].val[0] = Grid[idx000].m;
                gridCells[cellIndex].val[1] = Grid[idx001].m;
                gridCells[cellIndex].val[2] = Grid[idx010].m;
                gridCells[cellIndex].val[3] = Grid[idx011].m;
                gridCells[cellIndex].val[4] = Grid[idx100].m;
                gridCells[cellIndex].val[5] = Grid[idx101].m;
                gridCells[cellIndex].val[6] = Grid[idx110].m;
                gridCells[cellIndex].val[7] = Grid[idx111].m;

                cellIndex++;
            }
        }
    }
}


int main() {

    MaterialProperties materialProps;
    TimeProperties timeProps;
    SimulationProperties simProps;
    GridProperties gridProps;
    AnimationProperties animProps;
    PathsProperties pathProps;

    // Wczytaj dane z pliku config.json
    std::string configFilePath = "config.json";
    LoadConfiguration(configFilePath, materialProps, timeProps, simProps, gridProps, animProps, pathProps);

    size_t n_timesteps = (timeProps.t_end - timeProps.t_0) / timeProps.dt;

    // Obliczenie współczynników Lamego
    materialProps.calculateLameCoefficients();

    // Alokacja pamięci na GPU
    SimulationProperties* simProps_gpu;
    MaterialProperties* materialProps_gpu;
    GridProperties* gridProps_gpu;

    cudaMalloc((void**)&simProps_gpu, sizeof(SimulationProperties));
    cudaMalloc((void**)&materialProps_gpu, sizeof(MaterialProperties));
    cudaMalloc((void**)&gridProps_gpu, sizeof(GridProperties));
    
    // Kopiowanie danych na GPU
    cudaMemcpy(simProps_gpu, &simProps, sizeof(SimulationProperties), cudaMemcpyHostToDevice);
    cudaMemcpy(materialProps_gpu, &materialProps, sizeof(MaterialProperties), cudaMemcpyHostToDevice);
    cudaMemcpy(gridProps_gpu, &gridProps, sizeof(GridProperties), cudaMemcpyHostToDevice);

    // Wyświetlenie wyników
    std::cout << "Theta_c: " << materialProps.theta_c << std::endl;
    std::cout << "Theta_s: " << materialProps.theta_s << std::endl;
    std::cout << "xi: " << materialProps.xi << std::endl;
    std::cout << "rho_0: " << materialProps.rho_0 << std::endl;
    std::cout << "Modul Younga (E_0): " << materialProps.E_0 << std::endl;
    std::cout << "nu: " << materialProps.nu << std::endl;
    std::cout << "Współczynnik Lamego (mu_0): " << materialProps.mu_0 << std::endl;
    std::cout << "Współczynnik Lamego (lambda_0): " << materialProps.lambda_0 << std::endl;
    std::cout << "Krok czasowy (dt): " << timeProps.dt << std::endl;
    std::cout << "t: " << timeProps.t << std::endl;
    std::cout << "t_0: " << timeProps.t_0 << std::endl;
    std::cout << "t_end: " << timeProps.t_end << std::endl;
    std::cout << "sticky: " << simProps.sticky << std::endl;
    std::cout << "alpha: " << simProps.alpha << std::endl;
    std::cout << "particle_diameter: " << simProps.particle_diameter << std::endl;
    std::cout << "h: " << gridProps.hh << std::endl;
    std::cout << "MARCHING_CUBES_EPSILON: " << gridProps.MARCHING_CUBES_EPSILON << std::endl;
    std::cout << "isolevel: " << gridProps.isolevel << std::endl;
    std::cout << "step: " << animProps.step << std::endl; 
    std::cout << "base_path: " << pathProps.base_path << std::endl;
    std::cout << "result_path: " << pathProps.result_path << std::endl;
    std::cout << "result_name: " << pathProps.result_name << std::endl;
    std::cout << "walls_path: " << pathProps.walls_path << std::endl;
    std::cout << "walls_name: " << pathProps.walls_name << std::endl;

    // wczytanie siatki
    Mesh Walls;
    loadOBJ(pathProps.walls_path + pathProps.walls_name, Walls);

    // bounding box 
    h = gridProps.hh;
    bounding_box(Walls, x_min_grid, y_min_grid, z_min_grid, x_max_grid, y_max_grid, z_max_grid);

    x_min_grid -= h;
    y_min_grid -= h;
    z_min_grid -= h;

    x_max_grid += h;
    y_max_grid += h;
    z_max_grid += h;

    n_x_grid = static_cast<int>((x_max_grid - x_min_grid) / h);
    n_y_grid = static_cast<int>((y_max_grid - y_min_grid) / h);
    n_z_grid = static_cast<int>((z_max_grid - z_min_grid) / h);

    n_grid_nodes = n_x_grid * n_y_grid * n_z_grid;
    
    // przekopiowanie na gpu
    cudaMemcpyToSymbol(device_sticky, &simProps.sticky, sizeof(float));
    cudaMemcpyToSymbol(device_MCE, &gridProps.MARCHING_CUBES_EPSILON, sizeof(float));
    cudaMemcpyToSymbol(n_x_grid, &n_x_grid, sizeof(int));
    cudaMemcpyToSymbol(n_y_grid, &n_y_grid, sizeof(int));
    cudaMemcpyToSymbol(n_z_grid, &n_z_grid, sizeof(int));
    cudaMemcpyToSymbol(n_grid_nodes, &n_grid_nodes, sizeof(int));
    cudaMemcpyToSymbol(h, &h, sizeof(Precision));

    cudaMemcpyToSymbol(x_min_grid, &x_min_grid, sizeof(Precision));
    cudaMemcpyToSymbol(y_min_grid, &y_min_grid, sizeof(Precision));
    cudaMemcpyToSymbol(z_min_grid, &z_min_grid, sizeof(Precision));


    // do losowania pozycji
    std::random_device rd;
    std::mt19937 gen(rd()); 
    std::uniform_real_distribution<> dis(0.0, 1.0); // Przedział [0.0, 1.0)

    //inicjalizacja siatki
    std::vector<Node> Grid;
    Grid.reserve(n_grid_nodes);

    for (int i = 0; i < n_x_grid; i++) {
        for (int j = 0; j < n_y_grid; j++) {
            for (int k = 0; k < n_z_grid; k++) {
                //Grid.push_back(Node(i * h, j * h, k * h));
                Grid.emplace_back(Node(i * h + x_min_grid, j * h + y_min_grid, k * h + z_min_grid));
            }
        }
    }

    // SDF do kolizji
    std::vector<SDF_Node> SDF_grid;

    const int sdf_n_x_grid = n_x_grid;
    const int sdf_n_y_grid = n_y_grid;
    const int sdf_n_z_grid = n_z_grid;
    const int sdf_n_grid_nodes = sdf_n_x_grid * sdf_n_y_grid * sdf_n_z_grid;

    SDF_grid.reserve(sdf_n_grid_nodes);
    for (int i = 0; i < sdf_n_x_grid; i++) {
        for (int j = 0; j < sdf_n_y_grid; j++) {
            for (int k = 0; k < sdf_n_z_grid; k++) {
                SDF_grid.emplace_back(SDF_Node(i * h + x_min_grid, j * h + y_min_grid, k * h + z_min_grid));
            }
        }
    }



    // EMITERY
    // nizej opisalem, jak te emitery mniej wiecej miałyby dzialac, obecnie sa wylaczone
    //wczytanie powierzchni emiterow

    // to powinno byc tak:
    // w pliku wejsciowym
    // musi byc lista z sciazkami do plikow obj, majacych byc emiterami
    // oraz ich parametry: predkosc poczatkowa czatek (kierunek bedzie normalny do powierzchni zawsze), liczba czastek przypadajaca na dany emiter na 1 krok czasowy
    // gdzies pewnie potrzebna tez bedzie liczba tych emiterow (zeby wiedziec ile ich wczytac)

    // to tutaj jest tylko przykladowe, normalnie powinno byc zrobione przez wczytywanie z pliku
    //Mesh Emitter_mesh;
    //loadOBJ(emitter_path, Emitter_mesh);

    std::vector<Emitter> Emitters;
    // i tutaj powinna byc petla, wykonywana tyle razy ile ma byc emiterow
    // kolejne emitery maja kolejne wczytane dla nich siatki i parametry oraz wyznaczajaca, ktory indeks bedzie pierwszy
    //Emitters.push_back(Emitter(Emitter_mesh, 1.0, 100, 0));     // parametry: siatka, predkosc poczakowa, liczba czastek na krok czasowy, indeks pierwszej czestki
                                                                // co do tego indeksu, to jesli mamy 3 emitery, to pierwszy ma ten indeks = 0, drugi ma rowny indesowi ostatniej czastki pierwszego emitara, czyli 
                                                                // jesli pierwszy dodaje 100 czastek, to drugi musi starowac z indeksem = 100, podobnie kolejny

    std::vector<Eigen::Vector3i> emitters_triangle_vertices_indices;
    std::vector<Eigen::Vector3i> emitters_triangle_normals_indices;
    std::vector<Vector> emitters_vertices;
    std::vector<Vector> emitters_normals;


    // trzeba splaszczyc, zeby wyslac na cuda
    size_t n_emitter_vertices = 0;
    size_t n_emitter_normals = 0;
    size_t n_emitter_triangles = 0;
    for (int i = 0; i < Emitters.size(); i++) {
        Emitters[i].starting_triangle_index = n_emitter_triangles;
        Emitters[i].n_triangles = Emitters[i].mesh.triangles.size();
        Emitters[i].ending_triangle_index = n_emitter_triangles + Emitters[i].mesh.triangles.size();
        n_emitter_triangles += Emitters[i].mesh.triangles.size();
        for (int j = 0; j < Emitters[i].mesh.triangles.size(); j++) {
            emitters_triangle_vertices_indices.push_back(Eigen::Vector3i(Emitters[i].mesh.triangles[j].vertexIndices[0] + n_emitter_vertices, Emitters[i].mesh.triangles[j].vertexIndices[1] + n_emitter_vertices, Emitters[i].mesh.triangles[j].vertexIndices[2] + n_emitter_vertices)); // sprawdzic, czy dziala dla wiecej niz jednego
            emitters_triangle_normals_indices.push_back(Eigen::Vector3i(Emitters[i].mesh.triangles[j].normalIndices[0] + n_emitter_normals, Emitters[i].mesh.triangles[j].normalIndices[1] + n_emitter_normals, Emitters[i].mesh.triangles[j].normalIndices[2] + n_emitter_normals));
        }
        for (int j = 0; j < Emitters[i].mesh.vertices.size(); j++) {
            emitters_vertices.push_back(Emitters[i].mesh.vertices[j]);
        }

        for (int j = 0; j < Emitters[i].mesh.normals.size(); j++) {
            emitters_normals.push_back(Emitters[i].mesh.normals[j]);
        }

        n_emitter_vertices += Emitters[i].mesh.vertices.size();
        n_emitter_normals += Emitters[i].mesh.normals.size();
    }

    // ile czastek bedzie dodawane w danym kroku czasowym
    size_t emitters_particles_added_per_timestep = 0;
    for (int i = 0; i < Emitters.size(); i++) {
        emitters_particles_added_per_timestep += Emitters[i].particles_per_timestep;
    }

    // KONIEC EMITEROW


    // inicjalizacja gridu do marchingcubes
    // na razie bedzie tozsama z tym do obliczen
    // jakby miala byc inna, to trzeba zrobic cos na ksztalt particle_to_grid dla tej innej siatki

    const int MC_n_x_grid = n_x_grid - 1; // odejmujemy 1 bo dla mpm grid to wezly, a tutaj to beda komorki
    const int MC_n_y_grid = n_y_grid - 1;
    const int MC_n_z_grid = n_z_grid - 1;
    const int MC_n_grid_nodes = MC_n_x_grid * MC_n_y_grid * MC_n_z_grid;

    std::vector<GRIDCELL> MC_Grid;
    MC_Grid.reserve(MC_n_grid_nodes);

    TRIANGLE* triangles = new TRIANGLE[5 * MC_n_grid_nodes];
    //TRIANGLE triangles[5 * MC_n_grid_nodes]{};

    int totalTriangles = 0;

    for (int i = 0; i < MC_n_x_grid; i++) {
        for (int j = 0; j < MC_n_y_grid; j++) {
            for (int k = 0; k < MC_n_z_grid; k++) {
                GRIDCELL cell;

                int idx000 = i * n_y_grid * n_z_grid + j * n_z_grid + k;
                int idx001 = (i + 1) * n_y_grid * n_z_grid + j * n_z_grid + k;
                int idx010 = (i + 1) * n_y_grid * n_z_grid + (j + 1) * n_z_grid + k;
                int idx011 = i * n_y_grid * n_z_grid + (j + 1) * n_z_grid + k;
                int idx100 = i * n_y_grid * n_z_grid + j * n_z_grid + (k + 1);
                int idx101 = (i + 1) * n_y_grid * n_z_grid + j * n_z_grid + (k + 1);
                int idx110 = (i + 1) * n_y_grid * n_z_grid + (j + 1) * n_z_grid + (k + 1);
                int idx111 = i * n_y_grid * n_z_grid + (j + 1) * n_z_grid + (k + 1);

                // pozycje wierzcholkow
                cell.p[0] = { Grid[idx000].x.x(), Grid[idx000].x.y(), Grid[idx000].x.z() };
                cell.p[1] = { Grid[idx001].x.x(), Grid[idx001].x.y(), Grid[idx001].x.z() };
                cell.p[2] = { Grid[idx010].x.x(), Grid[idx010].x.y(), Grid[idx010].x.z() };
                cell.p[3] = { Grid[idx011].x.x(), Grid[idx011].x.y(), Grid[idx011].x.z() };
                cell.p[4] = { Grid[idx100].x.x(), Grid[idx100].x.y(), Grid[idx100].x.z() };
                cell.p[5] = { Grid[idx101].x.x(), Grid[idx101].x.y(), Grid[idx101].x.z() };
                cell.p[6] = { Grid[idx110].x.x(), Grid[idx110].x.y(), Grid[idx110].x.z() };
                cell.p[7] = { Grid[idx111].x.x(), Grid[idx111].x.y(), Grid[idx111].x.z() };

                // Wartosci w wierzcholkach (w tym miejscu to i tak 0), na razie zostaje dla testu, czy dziala
                cell.val[0] = Grid[idx000].m;
                cell.val[1] = Grid[idx001].m;
                cell.val[2] = Grid[idx010].m;
                cell.val[3] = Grid[idx011].m;
                cell.val[4] = Grid[idx100].m;
                cell.val[5] = Grid[idx101].m;
                cell.val[6] = Grid[idx110].m;
                cell.val[7] = Grid[idx111].m;

                MC_Grid.emplace_back(cell);
            }
        }
    }


    // wczytanie siatek majacych byc wypelnionych czastkami
    // 
    // ja to zrobie na razie prowizorycznie, zeby w miare wygodnie wczytac siatki
    // co jest wazne przy przygotowaniu tych siatek:
    // - musza znajdowac sie wewnatrz tej siatki ze sciankami
    // - musza miec normale skierowane do srodka
    // - nie moga na siebie nachodzic
    // 
    // nie ma kontroli bledow, wiec jak cos bedzie nie tak, to nie bedzie dzialac

    
    // UWAGA przy wczytywaniu jakichkolwiek siatek
    
    // siatka ze scianami:
    // 1. musi byc to pojedynczy plik
    // 2. musi być miec cos na ksztalt boundig box
    // 3. wektory normalne musza byc skierowane w strone obszaru symulacji
    // czyli na przyklad, jesli chcemy symulowac sytuacje w pudelku szeciennym, to:
    // 1. robimy siatke szescianu
    // 2. upewniamy sie, ze wektory normalne sa skierowane do srodka szescianu
    // 3. zapisujemy do pliku .obj
    
    // siatki do inicjalizacji czastek
    // 1. mozna wczytac wiele siatek, ale w pojedynczym pliku powinna byc jedna bryla
    // (w teorii jest szansa, ze dziala tez w innych przypadkach, jednak nie bylo to testowane)
    // 2. wektory normalne powinny byc skierowane w strone w ktorej maja sie pojawiac czastki,
    // czyli np. jesli maja sie pojawic w kuli, to wektory normalne musza byc skierowanie do
    // srodka kuli

    std::vector<std::string> snow_block_paths;
    for (int i = 0; i < pathProps.snow_objects_paths.size(); i++) {
        snow_block_paths.push_back(pathProps.snow_objects_paths[i] + pathProps.snow_objects_names[i]);
    }

    // predkosci w ruchu postepowym
    // kolejnosc jak powyzej przy wczytywaniu sciezek
    std::vector<Vector> snow_block_velocities;

    for (int i = 0; i < pathProps.snow_block_velocities.size(); i++) {
        snow_block_velocities.push_back(pathProps.snow_block_velocities[i]);
    }

    // predkosci obrotowe
    // kolejnosc jak przy wczytywaniu sciezek
    // jednostka to radiany na sekunde
    // przy wczytywaniu mozna dac np. obroty na sekunde, ale wtedy tutaj trzeba przeliczyc na radiany na sekunde
    std::vector<Vector> snow_block_rotation;
    for (int i = 0; i < pathProps.snow_block_rotation.size(); i++) {
        snow_block_rotation.push_back(pathProps.snow_block_rotation[i]);
    }

    // gestosci czastek w danym bloku sniegu
    // kolejnosc jak przy wczytywaniu sciezek
    std::vector<Precision> snow_block_rho;
    for (int i = 0; i < pathProps.snow_block_rho.size(); i++) {
        snow_block_rho.push_back(pathProps.snow_block_rho[i]);
    }

    // liczba siatek wypelnianych sniegiem do wczytania
    size_t n_snow_block_meshes = snow_block_paths.size();

    std::vector<Mesh> snow_block_meshes;

    for (size_t i = 0; i < n_snow_block_meshes; i++) {
        Mesh particles_block;
        loadOBJ(snow_block_paths[i], particles_block);
        snow_block_meshes.push_back(particles_block);
    }
    
    // -Z forward, Y up


    // wyznaczenie srednich wspolrzednych wierzcholkow (dla kul beda to ich srodki)
    // wykonujemy to zeby moc zadac predkosc obrotowa na poczatku symulacji
    std::vector<Vector> snow_block_centres;
    for (size_t i = 0; i < n_snow_block_meshes; i++) {

        Vector center = Vector::Zero();

        for (size_t j = 0; j < snow_block_meshes[i].vertices.size(); j++) {
            center += snow_block_meshes[i].vertices[j];
        }
        center /= snow_block_meshes[i].vertices.size();
        snow_block_centres.push_back(center);
    }

    //inicjalizacja czastek
    std::vector<Particle> Particles;


    // wyznaczamy objetosc wczytanej siatki (na zasadzie jak z twierdzenia Ostrogradskiego-Gaussa)
    // z objetosci liczbe czastek, ktora ja wypelni
    // i potem dla siatki liczymy sdf (wektory normalne do srodka, wiec odleglosc dodatnia jest w srodku)
    // jesli odlaglosc jest dodatnia, to wstawiamy czastke, inkrementujemy

    for (size_t i = 0; i < n_snow_block_meshes; i++) {
        Precision snow_block_volume = calculateVolume(snow_block_meshes[i]);
        size_t snow_block_particle_count = static_cast<size_t>(std::round(snow_block_volume / simProps.particle_volume));
        std::cout << "n: " << snow_block_particle_count << "\n";

        // wyznaczanie sdf dla bryl sniegu

        for (int i = 0; i < SDF_grid.size(); i++) {
            SDF_grid[i].signed_dist = std::numeric_limits<float>::max();
            SDF_grid[i].dist_vec = Vector::Zero();
        }

        

        for (int sdf_i = 0; sdf_i < SDF_grid.size(); sdf_i++) {
            for (int sdf_j = 0; sdf_j < snow_block_meshes[i].triangles.size(); sdf_j++) {
                // szukamy minimum dla danego i 
                ClosestPointResult sdf_point_triangle = closest_point_triangle(snow_block_meshes[i], sdf_j, SDF_grid[sdf_i]);
                Vector sdf_distance_vector = vector_distance(sdf_point_triangle.closestPoint, SDF_grid[sdf_i]);
                Precision sdf_distance = sdf_distance_vector.norm();
                if (std::fabs(SDF_grid[sdf_i].signed_dist) > sdf_distance) {
                    SDF_grid[sdf_i].dist_vec = sdf_distance_vector;
                    Precision sign = sdf_distance_vector.transpose() * sdf_point_triangle.normal;
                    SDF_grid[sdf_i].signed_dist = sdf_distance;
                    if (sign > 0.0) {
                        SDF_grid[sdf_i].signed_dist *= (-1);
                    }
                }
            }
        }
        size_t j_particles = 0;
        while (j_particles < snow_block_particle_count) {

            Vector particle_pos_drawn = Vector((x_max_grid - x_min_grid - 4 * h) * dis(gen) + x_min_grid + 2 * h, (y_max_grid - y_min_grid - 4 * h) * dis(gen) + y_min_grid + 2 * h, (z_max_grid - z_min_grid - 4 * h) * dis(gen) + z_min_grid + 2 * h);

            if (signed_distance_interpolation(particle_pos_drawn, SDF_grid.data()) > 0) {
                // pierwsze 3 argumenty w konstruktorze Particle to pozycja, nastepne 3 to predkosc, dalej gestosc poczatkowa i objetosc czastki
                // predkosc jest suma predkosci w ruchu prostoliniowym i predkosci obrotowej (ruch plaski)

                // skladowe od predkosci obrotowej wyliczamy na podstawie wlasnie tych predkosci obrotowych
                Vector omega = snow_block_rotation[i];
                Vector velociy_from_rotation = omega.cross(particle_pos_drawn - snow_block_centres[i]);
                Vector velocity_total = velociy_from_rotation + snow_block_velocities[i];

                Particles.push_back(Particle(particle_pos_drawn.x(), particle_pos_drawn.y(), particle_pos_drawn.z(), velocity_total.x(), velocity_total.y(), velocity_total.z(), snow_block_rho[i], simProps.particle_volume));
                j_particles++;
            }
        }

    }
    size_t n_particles = Particles.size();

    // tutaj sa dodawane czastki, ktore wyleca z emiterow; wszystkie czastki, ktore wezma udzial w symulacji sa juz tutaj przewidziane, zeby nie ptrzeba bylo zmieniac zalokowanej pamieci
    
    for (int p = 0; p < emitters_particles_added_per_timestep * n_timesteps; p++) {
        Particles.push_back(Particle(std::numeric_limits<Precision>::max(), std::numeric_limits<Precision>::max(), std::numeric_limits<Precision>::max(), 0, 0, 0, 0, 0));
    }


    // wyznaczyc sdf
    // jesli bedzie animacja (ruch siatki), to dla kazdej klatki trzeba liczyc na nowo i wyznaczac sprytna interpolacje pomiedzy klatkami
    // do tego w kolizjach trzeba uwzglednic predkosc siatki

    // na razie zrobione w sposob naiwny
    // w bardziej zlozonych przypadkach bedzie sie liczylo za dlugo
    // wiem, jak znacznie to przypieszyc, moze w przyszlosci napisze

    // najpierw reset
    for (int i = 0; i < SDF_grid.size(); i++) {
        SDF_grid[i].signed_dist = std::numeric_limits<float>::max();
        SDF_grid[i].dist_vec = Vector::Zero();
    }

    for (int i = 0; i < SDF_grid.size(); i++) {
        for (int j = 0; j < Walls.triangles.size(); j++) {
            // szukamy minimum dla danego i 
            ClosestPointResult sdf_point_triangle = closest_point_triangle(Walls, j, SDF_grid[i]);
            Vector sdf_distance_vector = vector_distance(sdf_point_triangle.closestPoint, SDF_grid[i]);
            Precision sdf_distance = sdf_distance_vector.norm();
            if (std::fabs(SDF_grid[i].signed_dist) > sdf_distance) {
                SDF_grid[i].dist_vec = sdf_distance_vector;
                Precision sign = sdf_distance_vector.transpose() * sdf_point_triangle.normal;
                SDF_grid[i].signed_dist = sdf_distance;
                if (sign > 0.0) {
                    SDF_grid[i].signed_dist *= (-1);
                }
            }
        }
    }


    // glowna petla
    Precision weight; // wynik funkcji wagowej z interpolacji
    Precision rhop; // gestosc czastki

    // przygotowanie miejsca na karcie graficzniej
    // jesli sa emitery, to miejsce na wszystkie czastki, ktore wyemituja w trakcie symulacji jest alokowane już na poczatku
    Particle* particles_gpu;
    size_t particles_size_malloc = Particles.size() * sizeof(Particle);
    size_t particles_size = Particles.size() * sizeof(Particle);

    Node* grid_gpu;
    size_t grid_size = n_grid_nodes * sizeof(Node);

    SDF_Node* sdf_grid_gpu;
    size_t sdf_grid_size = sdf_n_grid_nodes * sizeof(SDF_Node);

    Emitter* emitters_gpu;
    size_t n_emitters_malloc = Emitters.size() * sizeof(Emitter);

    Eigen::Vector3i* emitters_triangles_vertex_gpu;
    size_t n_emitter_triangles_vertex_malloc = emitters_triangle_vertices_indices.size() * sizeof(Eigen::Vector3i);

    Eigen::Vector3i* emitters_triangles_normal_gpu;
    size_t n_emitter_triangles_normal_malloc = emitters_triangle_normals_indices.size() * sizeof(Eigen::Vector3i);

    Vector* emitter_vertices;
    size_t n_emitter_vertices_malloc = emitters_vertices.size() * sizeof(Vector);

    Vector* emitter_normals;


    
    cudaMalloc((void**)&particles_gpu, particles_size_malloc);
    cudaMalloc((void**)&grid_gpu, grid_size);
    cudaMalloc((void**)&sdf_grid_gpu, sdf_grid_size);

    cudaMalloc((void**)&emitters_gpu, n_emitters_malloc);
    cudaMalloc((void**)&emitters_triangles_vertex_gpu, n_emitter_triangles_vertex_malloc);
    cudaMalloc((void**)&emitters_triangles_normal_gpu, n_emitter_triangles_normal_malloc);

    cudaMalloc((void**)&emitter_vertices, n_emitter_vertices_malloc);
    cudaMalloc((void**)&emitter_normals, n_emitter_vertices_malloc);

    cudaMemcpy(particles_gpu, Particles.data(), particles_size, cudaMemcpyHostToDevice);
    cudaMemcpy(grid_gpu, Grid.data(), grid_size, cudaMemcpyHostToDevice);
    cudaMemcpy(sdf_grid_gpu, SDF_grid.data(), sdf_grid_size, cudaMemcpyHostToDevice);

    cudaMemcpy(emitters_gpu, Emitters.data(), n_emitters_malloc, cudaMemcpyHostToDevice);
    cudaMemcpy(emitters_triangles_vertex_gpu, emitters_triangle_vertices_indices.data(), n_emitter_triangles_vertex_malloc, cudaMemcpyHostToDevice);
    cudaMemcpy(emitters_triangles_normal_gpu, emitters_triangle_normals_indices.data(), n_emitter_triangles_normal_malloc, cudaMemcpyHostToDevice);

    cudaMemcpy(emitter_vertices, emitters_vertices.data(), n_emitter_vertices_malloc, cudaMemcpyHostToDevice);
    cudaMemcpy(emitter_normals, emitters_normals.data(), n_emitter_vertices_malloc, cudaMemcpyHostToDevice);

    
    // mozna by sie zastanowic, czy nie powinno byc inaczej
    int threadsPerBlock = 512;
    int blocksPerGrid = (Grid.size() + threadsPerBlock - 1) / threadsPerBlock;
    int blocksPerParticles = (Particles.size() + threadsPerBlock - 1) / threadsPerBlock;

    
    // interpolacja z czastek na siatke, pierwsza poza petla, zeby wyznaczyc objetosci (punkt 2 algorytmu)
    particle_to_grid_mass << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles);
    cudaDeviceSynchronize();

    calculate_V << <blocksPerParticles, threadsPerBlock >> > (grid_gpu, particles_gpu, n_grid_nodes, n_particles);
    cudaDeviceSynchronize();

    
    // do wizualizacji
    std::vector<Precision> x_particle_data;
    x_particle_data.reserve(Particles.size());
    std::vector<Precision> y_particle_data;
    y_particle_data.reserve(Particles.size());
    std::vector<Precision> z_particle_data;
    z_particle_data.reserve(Particles.size());

    int iteration = 0;
    int frame = 0;
    //glTranslatef(-0.5f, -0.5f, -0.5);

    // mozna sie zastanowic, czy nie lepiej for, jako ze t bedzie obarczone bledem numerycznym
    // albo dac maksymalna liczbe iteracji

    size_t emitter_seed = 25519;

    while (timeProps.t < timeProps.t_end) { //&& !glfwWindowShouldClose(window)) {

        // krok w mpm        
        MPM_step(particles_gpu, grid_gpu, sdf_grid_gpu, n_particles, n_grid_nodes, timeProps.dt, blocksPerGrid, blocksPerParticles, threadsPerBlock, emitters_gpu, Emitters.size(), emitters_triangles_vertex_gpu, emitters_triangles_normal_gpu, n_emitter_triangles, emitter_vertices, n_emitter_vertices, emitter_normals, emitter_seed, emitters_particles_added_per_timestep, materialProps_gpu, simProps_gpu);

        // animacja
        // co ile krokow czasowych zapisac wynik
        if (iteration % animProps.step == 0) {
            cudaMemcpy(Grid.data(), grid_gpu, grid_size, cudaMemcpyDeviceToHost);

            // aktualizacja struktury do marching cubes
            update_MC_Grid_Cell_Values(Grid, MC_Grid, MC_n_x_grid, MC_n_y_grid, MC_n_z_grid);

            // Polygonise
            cudaError_t cudaStatus_anim = polygoniseWithCuda(MC_Grid.data(), triangles, MC_n_grid_nodes, gridProps.isolevel, &totalTriangles);
            if (cudaStatus_anim != cudaSuccess) {
                fprintf(stderr, "polygoniseWithCuda failed!");
                return 1;
            }

            // Saving triangles to obj file
            // mozna by dodac wiodace zera
            std::cout << pathProps.result_path << "\n";
            std::string filename = pathProps.result_path + pathProps.result_name + std::to_string(frame) + ".obj";
            std::ofstream outFile(filename);

            if (!outFile.is_open()) {
                std::cerr << "Error opening file" << std::endl;
                return 1;
            }

            // vertices
            for (int i = 0; i < totalTriangles; i++) {
                for (int j = 0; j < 3; j++) {
                    outFile << "v " << triangles[i].p[j].x << " " << triangles[i].p[j].y << " " << triangles[i].p[j].z << std::endl;
                }
            }

            // faces
            for (int i = 0; i < totalTriangles; i++) {
                outFile << "f " << 1 + i * 3 << " " << 2 + i * 3 << " " << 3 + i * 3 << std::endl;
            }

            outFile.close();

            frame++;

        }

        // do sprawdzania, czy masa sie zgadza
        Precision sum_mass = 0;
        Precision sum_momentum = 0;
        for (int i = 0; i < Particles.size(); i++) {
            sum_mass += Particles[i].m;
        }


        timeProps.t += timeProps.dt;
        std::cout << "=======================================\n";
        std::cout << "Timestep number: " << iteration << "\n";
        std::cout << "Simulation time: " << timeProps.t << "\n";
        std::cout << "Total mass in simulation: " << sum_mass << "\n";

        iteration++;
    }

    cudaMemcpy(Particles.data(), particles_gpu, particles_size, cudaMemcpyDeviceToHost);
    cudaMemcpy(Grid.data(), grid_gpu, grid_size, cudaMemcpyDeviceToHost);

    cudaFree(particles_gpu);
    cudaFree(grid_gpu);
    cudaFree(sdf_grid_gpu);
    cudaFree(materialProps_gpu);
    cudaFree(simProps_gpu);
    cudaFree(gridProps_gpu);

    delete triangles;

    return 0;
}
